/*
 File        : menu_memory.c

 Author      : Sergiy Uvarov - Copyright (C) 2001

 Description : This is the file to create windows of memory.

 Copyright notice:

 avr_simulator - A GNU/Linux simulator for the Atmel AVR series
 of microcontrollers. Copyright (C) 2001 Sergiy Uvarov

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 Sergiy Uvarov

 E-mail : colonel@ff.dsu.dp.ua
*/

#include <stdio.h>
#include <inttypes.h>
#include <stdlib.h>
#include <libintl.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <Xm/XmAll.h>
#include "types.h"
#include "global.h"
#include "menu_memory.h"

#define _(x) gettext(x)
#define N_(x) (x)
#define MAX_SRAM_LINES    20
#define MAX_EEPROM_LINES  20
#define MAX_IO_PORT_LINES 30

int type_gpr_window=HEX,type_sram_window=HEX,
  type_eeprom_window=HEX,
  address_sram=0,          /* address of start point of sram_window      */
  nlines_sram=8,           /* number of lines in sram_window             */
  nlines_sram_old=8,       /* need only for tip window in sram_window    */
  nlines_io=10,            /* number of lines in io_window               */
  nlines_eeprom=4,         /* number of lines in eeprom_window           */
  nlines_eeprom_old=4,     /* need only for tip window in eeprom_window  */
  address_eeprom=0,        /* address in eeprom_window                   */
  address_io=0,            /* address of start point of io_window        */
  address_io_old=0,         /* need only for tip window in io_window      */
  type_micro_old,          /* need only for tip windows                  */    
  height_min_io_window,    /* height of io_window without port lines     */
  height_min_sram_window,  /* height of sram_window without data lines   */
  height_min_eeprom_window,/* height of eeprom_window without data lines */
  eeprom_window_step,      /* number of incr. pixels of eeprom_window    */
  sram_window_step,        /* number of incr. pixels of sram_window      */
  io_window_step;          /* number of incr. pixels of io_window        */
  
Widget *field=NULL,        /* edit field in gpr_window                */
  *field_sram=NULL,        /* edit field in sram_window               */
  vscr_sram_bar,           /* scrollbar in sram_window                */
  *label_sram=NULL,        /* array of labels(address) in sram_window */
  vscr_io_bar,             /* scrollbar in io_window                  */
  *label_io=NULL,          /* array of labels in io_window            */
  **bit_button=NULL,       /* array of bit buttons                    */
  *label_io_val=NULL,      /* array of values in io_window            */
  *label_io_addr=NULL,     /* aray of address labels on io_window     */
  dataform_io,             /* form io_widget to place data            */
  dataform_sram,           /* form sram_widget to place data          */
  dataform_eeprom,         /* form eeprom_widget to place data        */
  *field_eeprom=NULL,      /* edit field in eeprom_window             */
  vscr_eeprom_bar,         /* scrollbar in eeprom_window              */
  *label_eeprom=NULL;      /* array of labels(address) in eeprom_window */

char **gpr_tip_str=NULL;
char **sram_tip_str=NULL;
char **eeprom_tip_str=NULL;

Widget gpr_window,io_window,sram_window,eeprom_window,memory_window=0;

char *memory_type_str[]={
  N_("Registers"),N_("I/O registers"),N_("Ext I/O Registers"),N_("SRAM"),
  N_("Data"),N_("EEPROM")};

void update_gpr_window();
void update_sram_window();
void update_eeprom_window();
void update_io_window();
void full_update_io_window();
void full_update_sram_window();
void full_update_eeprom_window();
void switch_button_gpr();
void switch_button_sram();
void switch_button_eeprom();
void close_button_gpr();
void close_button_sram();
void close_button_io();
void close_button_eeprom();
void scroll_bar_io_callback();
void scroll_bar_sram_callback();
void scroll_bar_eeprom_callback();
void get_focus_field_callback();
void get_focus_field_sram_callback();
void get_focus_field_eeprom_callback();
void lose_focus_field_callback();
void lose_focus_field_sram_callback();
void lose_focus_field_eeprom_callback();
void field_changed_callback(); 
void field_sram_changed_callback();
void field_eeprom_changed_callback();
void set_size_sram_window();
void set_size_io_window();
void set_size_eeprom_window();
void bit_button_callback();
void resize_io();
void resize_sram();
void resize_eeprom();
void create_io_ports_line();
void create_sram_line();
void create_eeprom_line();

void init_windows_memory(void)
{
  int n;

  /* allocate memory for 'mem_file_changed' */
  if(mem_file_changed!=NULL) free(mem_file_changed);
  size_mem_file=(controllers[type_micro].size_gpr+
		 controllers[type_micro].size_io_reg+
		 controllers[type_micro].size_sram)/8+1;
  
  if((mem_file_changed=malloc(size_mem_file))==NULL)
    {
      puts("init_windows_memory(): can't allocate memory for 'mem_file_changed'");
      exit(1);
     }
  memset(mem_file_changed,0,size_mem_file);

  /* allocate memory for 'mem_file_changed' */
  if(eeprom_changed!=NULL) free(eeprom_changed);
  size_eeprom=controllers[type_micro].size_eeprom;
  
  if((eeprom_changed=malloc(size_eeprom))==NULL)
    {
      puts("init_windows_memory(): can't allocate memory for 'eeprom_changed'");
      exit(1);
    }
  memset(eeprom_changed,0,size_eeprom);
  
  /* set parameters of scrollbar of eeprom window */
  if(controllers[type_micro].size_sram)
    {
      XtVaSetValues(vscr_sram_bar,XmNsliderSize,nlines_sram,
		    XmNmaximum,controllers[type_micro].size_sram/16,
		    XmNvalue,0,NULL);
      address_sram=0;
    }
  else
    XtUnmanageChild(sram_window);
  
  /* set parameters of scrollbar of eeprom window */
  XtVaSetValues(vscr_eeprom_bar,XmNsliderSize,nlines_eeprom,
		XmNmaximum,controllers[type_micro].size_eeprom/16,
		XmNvalue,0,NULL);
  address_eeprom=0;

  /* set parameters of scrollbar of io window */
  for(n=0;controllers[type_micro].io_registers[n].code!=-1;n++);
  XtVaSetValues(vscr_io_bar,XmNmaximum,n,XmNsliderSize,nlines_io,NULL);
}

char *conv_to_bin(int number,char *buf)
{
  int n;
  
  if(buf==NULL)
    {
      puts("conv_to_bin(): buf==NULL");
      return NULL;
    }
  for(n=0;n<9;n++)
    {
      if(n==4) {buf[n]=' ';n++;}
      if(number%2)
	buf[8-n]='1';
      else
	buf[8-n]='0';
      number/=2;
    }
  buf[n]=0;
  return buf;
}

/* GPR window */
void create_gpr_window(void)
{
  
  char buf[5];
  Widget *label,button,shell,frame,gpr_form;
  int n,num,m,max=controllers[type_micro].size_gpr;
  XmString string;
  Arg args[19];
  
  if((label=malloc(max*sizeof(Widget)))==NULL)
    {
      puts("create_gpr_window(): can't allocate memory for 'label'");
      exit(1);
    }
  if((field=malloc(max*sizeof(Widget)))==NULL)
    {
      puts("create_gpr_window(): can't allocate memory for 'field'");
      exit(1);
    } 
  if((gpr_tip_str=malloc(max*sizeof(char *)))==NULL)
    {
      puts("create_gpr_window(): can't allocate memory for 'gpr_tip_str'");
      exit(1);
    }
  
  for(n=0;n<max;n++)
    {
      if((gpr_tip_str[n]=malloc(10*sizeof(char)))==NULL)
	{
	  printf("create_gpr_window(): \
can't allocate memory for 'gpr_tip_str[%d]'",n);
	  exit(1);
	}
      gpr_tip_str[n][0]=0;
    }


  shell=XtVaCreateWidget("Shell",xmDialogShellWidgetClass,toplevel,
			 XmNtitle,_("General Registers"),NULL);
  gpr_window=XtVaCreateWidget("form",xmFormWidgetClass,shell,
			      XmNautoUnmanage,False,
			      XmNdialogStyle,XmDIALOG_MODELESS,
			      XmNresizePolicy,XmRESIZE_NONE,
			      XmNfractionBase,5,NULL);
  
  frame=XtVaCreateManagedWidget("frame",xmFrameWidgetClass,gpr_window,
				XmNtopAttachment,XmATTACH_FORM,
				XmNtopOffset,5,
				XmNleftAttachment,XmATTACH_FORM,
				XmNleftOffset,5,
				XmNrightAttachment,XmATTACH_FORM,
				XmNrightOffset,5,NULL);
  gpr_form=XtVaCreateManagedWidget("form",xmFormWidgetClass,frame,NULL);

  XtSetArg(args[m=0],XmNleftAttachment,XmATTACH_WIDGET);m++;
  XtSetArg(args[m],XmNtopOffset,6);m++;
  XtSetArg(args[m],XmNleftOffset,0);m++;
  XtSetArg(args[m],XmNcolumns,3);m++;
  XtSetArg(args[m],XmNmaxLength,2);m++;
  XtSetArg(args[m],XmNvalue,"0");m++;
  XtSetArg(args[m],XmNcursorPositionVisible,False);m++;
  
  for(n=0;n<max;n++)
    {
      num=m;
      sprintf(buf,"r%-2d:",n);
      string=XmStringGenerate(buf,XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      label[n]=XtVaCreateManagedWidget("r",xmLabelGadgetClass,gpr_form,
				       XmNlabelString,string,
				       XmNtopOffset,8,XmNleftOffset,6,NULL);
      XmStringFree(string);

      XtSetArg(args[num],XmNleftWidget,label[n]);num++;
      
      if(n%4==0) 
	XtVaSetValues(label[n],XmNleftAttachment,XmATTACH_FORM,NULL);
      else
	XtVaSetValues(label[n],XmNleftAttachment,XmATTACH_WIDGET,
		      XmNleftWidget,field[n-1],NULL);
      if((n+1)%4==0)
	{
	  XtSetArg(args[num],XmNrightAttachment,XmATTACH_FORM);num++;
	  XtSetArg(args[num],XmNrightOffset,6);num++;
	}
	
      if(n<4)
	{
	  XtSetArg(args[num],XmNtopAttachment,XmATTACH_FORM);num++;
	  XtVaSetValues(label[n],XmNtopAttachment,XmATTACH_FORM,NULL);
	}
      else
	{
	  XtSetArg(args[num],XmNtopAttachment,XmATTACH_WIDGET);num++;
	  XtSetArg(args[num],XmNtopWidget,field[n-4]);num++;
	  XtVaSetValues(label[n],XmNtopAttachment,XmATTACH_WIDGET,
			XmNtopWidget,field[n-4],NULL);
	}

      if(max-n<5)
	{
	  XtSetArg(args[num],XmNbottomAttachment,XmATTACH_FORM);num++;
	  XtSetArg(args[num],XmNbottomOffset,6);num++;
	}
	
      field[n]=XtCreateManagedWidget("field",xmTextFieldWidgetClass,gpr_form,
				     args,num);
      xs_widget_tip(field[n],gpr_tip_str[n]);
      XtAddCallback(field[n],XmNfocusCallback,(XtCallbackProc)
		    get_focus_field_callback,NULL);
      XtAddCallback(field[n],XmNlosingFocusCallback,(XtCallbackProc)
		    lose_focus_field_callback,NULL);
    }

  string=XmStringGenerate(_("Decimal"),XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  button=XtVaCreateManagedWidget("Dec",xmPushButtonGadgetClass,
				 gpr_window,XmNtopAttachment,XmATTACH_WIDGET,
				 XmNtopOffset,6,XmNtopWidget,frame,
				 XmNleftAttachment,XmATTACH_POSITION,
				 XmNleftPosition,1,XmNlabelString,string,
				 XmNrightAttachment,XmATTACH_POSITION,
				 XmNrightPosition,2,
				 XmNbottomAttachment,XmATTACH_FORM,
				 XmNbottomOffset,6,NULL);
  
  XtAddCallback(button,
		XmNactivateCallback,(XtCallbackProc)switch_button_gpr,NULL);
  XmStringFree(string);
  
  string=XmStringGenerate(_("Close"),XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  button=XtVaCreateManagedWidget("Close",xmPushButtonGadgetClass,
				 gpr_window,XmNtopAttachment,XmATTACH_WIDGET,
				 XmNtopOffset,6,XmNtopWidget,frame,
				 XmNleftAttachment,XmATTACH_POSITION,
				 XmNleftPosition,3,
				 XmNrightAttachment,XmATTACH_POSITION,
				 XmNrightPosition,4,XmNlabelString,string,
				 XmNbottomAttachment,XmATTACH_FORM,
				 XmNbottomOffset,6,NULL);
  
  XtAddCallback(button,
		XmNactivateCallback,(XtCallbackProc)close_button_gpr,NULL);
  XmStringFree(string);
  
  XtAddCallback(shell,XmNpopupCallback,(XtCallbackProc)set_size_static_window,
		NULL);
  free(label);
}

void update_gpr_window(void)
{
  int n;
  char buf[4];
  unsigned char value;
    
  XtVaGetValues(menu[MEMORY_MENU].sub_menu[GPR_BUTTON].button,
		XmNset,&value,NULL);
  if(value!=XmSET) return;
  
  for(n=0;n<controllers[type_micro].size_gpr;n++)
    {
      if(mem_file_changed[n/8]&(1<<n%8))
	{
	  if(type_gpr_window==HEX)
	    sprintf(buf,"%X",gpr_pointer[n]);
	  else sprintf(buf,"%u",gpr_pointer[n]);
	  XtVaSetValues(field[n],XmNvalue,buf,NULL);
	  delete_tip(field[n],gpr_tip_str[n]);
	  xs_widget_tip(field[n],
			conv_to_bin((int)gpr_pointer[n],gpr_tip_str[n]));
	}
    }
}

void switch_button_gpr(Widget w,XtPointer client_data,XtPointer call_data)
{
  XmString string;
  int n;
  char buf[4];

  if(type_gpr_window==HEX) 
    {
      type_gpr_window=DEC;
      string=XmStringGenerate(_("Hex"),XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(w,XmNlabelString,string,NULL);
      XmStringFree(string);
      for(n=0;n<controllers[type_micro].size_gpr;n++)
	XtVaSetValues(field[n],XmNmaxLength,3,NULL);
    }
  else
    {
      type_gpr_window=HEX;
      string=XmStringGenerate(_("Decimal"),XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(w,XmNlabelString,string,NULL);
      XmStringFree(string);
      for(n=0;n<controllers[type_micro].size_gpr;n++)
	XtVaSetValues(field[n],XmNmaxLength,2,NULL);
    }
  
  /* update */
  for(n=0;n<controllers[type_micro].size_gpr;n++)
    {
      if(type_gpr_window==HEX)
	sprintf(buf,"%X",gpr_pointer[n]);
      else sprintf(buf,"%u",gpr_pointer[n]);
      XtVaSetValues(field[n],XmNvalue,buf,NULL); 
    }
}

void field_changed_callback(Widget w,XtPointer client_data,
			    XtPointer call_data)
{
  char *buf,buf2[10];
  int n;

  buf=XmTextFieldGetString(w);
  n=0;
  while(n<controllers[type_micro].size_gpr&&w!=field[n]) n++;
  if(type_gpr_window==DEC)
    gpr_pointer[n]=strtoul(buf,NULL,10);
  else
    {
      sprintf(buf2,"0x%s",buf);
      gpr_pointer[n]=strtoul(buf2,NULL,16);
    }
  delete_tip(field[n],gpr_tip_str[n]);
  xs_widget_tip(field[n],conv_to_bin((int)gpr_pointer[n],gpr_tip_str[n]));
  XtFree(buf);
  if(n>25) update_main(); 
}

void close_button_gpr(Widget w,XtPointer client_data,XtPointer call_data)
{
  XtVaSetValues(menu[MEMORY_MENU].sub_menu[GPR_BUTTON].button,XmNset,XmUNSET,
		NULL);
  XtUnmanageChild(gpr_window);
}

/* allow visible cursor */
void get_focus_field_callback (Widget w,XtPointer client_data,
			       XtPointer call_data)
{
  XtVaSetValues(w,XmNcursorPositionVisible,True,NULL);
  XtAddCallback(w,XmNvalueChangedCallback,(XtCallbackProc)
		field_changed_callback,NULL);
}

/* don't allow visible cursor*/
void lose_focus_field_callback (Widget w,XtPointer client_data,
			       XtPointer call_data)
{
  XtVaSetValues(w,XmNcursorPositionVisible,False,NULL);
  XtRemoveCallback(w,XmNvalueChangedCallback,(XtCallbackProc)
		   field_changed_callback,NULL);
}

/* SRAM window */
void create_sram_window(void)
{
  Widget scroll,sram_form,temp,shell,frame,*widgetlist;
  XmString string;
  int n,address,n_wid,t,n_args;
  char buf[6];
  Arg args[20];

  if((field_sram=malloc(16*MAX_SRAM_LINES*sizeof(Widget)))==NULL)
    {
      puts("create_sram_window(): can't allocate memory for 'field_sram'");
      exit(1);
    }
  if((label_sram=malloc(MAX_SRAM_LINES*sizeof(Widget)))==NULL)
    {
      puts("create_sram_window(): can't allocate memory for 'label_sram'");
      exit(1);
    }
  if((sram_tip_str=malloc(MAX_SRAM_LINES*16*sizeof(char *)))==NULL)
    {
      puts("create_sram_window(): can't allocate memory for 'sram_tip_str'");
      exit(1);
    }
  
  for(n=0;n<MAX_SRAM_LINES*16;n++)
    {
      if((sram_tip_str[n]=malloc(10*sizeof(char)))==NULL)
	{
	  printf("create_sram_window(): \
can't allocate memory for 'sram_tip_str[%d]'",n);
	  exit(1);
	}
      sram_tip_str[n][0]=0;
    }
  

  memset(label_sram,0,MAX_SRAM_LINES*sizeof(Widget));
  

  shell=XtVaCreateWidget("Shell",xmDialogShellWidgetClass,toplevel,
			 XmNtitle,_("SRAM"),NULL);
  sram_window=XtVaCreateWidget("form",xmFormWidgetClass,shell,
			       XmNautoUnmanage,False,
			       XmNdialogStyle,XmDIALOG_MODELESS,
			       XmNfractionBase,5,NULL);

  frame=XtVaCreateManagedWidget("frame",xmFrameWidgetClass,sram_window,
				XmNtopAttachment,XmATTACH_FORM,
				XmNtopOffset,5,
				XmNleftAttachment,XmATTACH_FORM,
				XmNleftOffset,5,
				XmNrightAttachment,XmATTACH_FORM,
				XmNrightOffset,5,NULL);
  sram_form=XtVaCreateManagedWidget("form",xmFormWidgetClass,frame,NULL);
  
  /* create firs row, labels */
  n_args=0;
  XtSetArg(args[n_args],XmNcursorPositionVisible,False);n_args++;
  XtSetArg(args[n_args],XmNeditable,False);n_args++;
  XtSetArg(args[n_args],XmNtopOffset,8);n_args++;
  XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_FORM);n_args++;
  t=n_args;

  XtSetArg(args[n_args],XmNleftOffset,3);n_args++;
  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_FORM);n_args++;
  XtSetArg(args[n_args],XmNvalue,"");n_args++;
  XtSetArg(args[n_args],XmNcolumns,4);n_args++;
  XtSetArg(args[n_args],XmNmaxLength,4);n_args++;

  temp=XtCreateManagedWidget("memory_label",xmTextFieldWidgetClass,sram_form,
			     args,n_args);

  n_args=t;
  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_WIDGET);n_args++;
  XtSetArg(args[n_args],XmNcolumns,3);n_args++;
  XtSetArg(args[n_args],XmNmaxLength,3);n_args++;
  XtSetArg(args[n_args],XmNleftOffset,2);n_args++;
  t=n_args;
  for(n=0;n<16;n++)
    {
      n_args=t;
      sprintf(buf,"%X",n);
      fill_zero(buf,2);
      XtSetArg(args[n_args],XmNvalue,buf);n_args++;
      XtSetArg(args[n_args],XmNleftWidget,temp);n_args++;
      temp=XtCreateManagedWidget("memory_label",xmTextFieldWidgetClass,
				 sram_form,args,n_args);
    }

  /* create scroll window */
  scroll=XtVaCreateManagedWidget("Scroll",xmScrolledWindowWidgetClass,
				 sram_form,
				 XmNscrollingPolicy,XmAPPLICATION_DEFINED,
				 XmNscrollBarDisplayPolicy,XmSTATIC,
				 XmNleftAttachment,XmATTACH_FORM,
				 XmNleftOffset,3,
				 XmNrightAttachment,XmATTACH_FORM,
				 XmNrightOffset,2,
				 XmNtopAttachment,XmATTACH_WIDGET,
				 XmNtopWidget,temp,
				 XmNbottomAttachment,XmATTACH_FORM,
				 XmNbottomOffset,3,
				 NULL);
  
  dataform_sram=XtVaCreateManagedWidget("form area",xmFormWidgetClass,scroll,
					NULL);
  
  vscr_sram_bar=XtVaCreateManagedWidget("scr_bar",xmScrollBarWidgetClass,
					scroll,XmNorientation,XmVERTICAL,
					XmNincrement,1,XmNminimum,0,
					XmNvalue,0,NULL);
  XtAddCallback(vscr_sram_bar,XmNvalueChangedCallback,
		(XtCallbackProc)scroll_bar_sram_callback,NULL);
  XtAddCallback(vscr_sram_bar,XmNdragCallback,
		(XtCallbackProc)scroll_bar_sram_callback,NULL);
  XmScrolledWindowSetAreas(scroll,NULL,vscr_sram_bar,dataform_sram);

  /* size of 'vscr_sram_bar' must be equal size of end space */
  XtVaGetValues(vscr_sram_bar,XmNwidth,(Dimension *)&n,NULL);
  XtVaSetValues(temp,XmNrightAttachment,XmATTACH_FORM,
		XmNrightOffset,n,NULL);
  
  /* create input fields */
  if((widgetlist=malloc(nlines_sram*17*sizeof(Widget)))==NULL)
    {
      puts("create_sram_window(): can't allocate memory for 'widgetlist'");
      exit(1);
    }
  n_wid=0;
  for(n=0,address=address_sram;n<nlines_sram;n++,address+=16)
    create_sram_line(n,address,widgetlist,&n_wid);
/*    XtManageChildren(widgetlist,n_wid); */
  free(widgetlist);

  string=XmStringGenerate(_("Decimal"),XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);  
  temp=XtVaCreateManagedWidget("Dec",xmPushButtonGadgetClass,
			       sram_window,
			       XmNleftAttachment,XmATTACH_POSITION,
			       XmNleftPosition,1,XmNlabelString,string,
			       XmNrightAttachment,XmATTACH_POSITION,
			       XmNrightPosition,2,
			       XmNbottomAttachment,XmATTACH_FORM,
			       XmNbottomOffset,6,NULL);
  
  XtAddCallback(temp,
		XmNactivateCallback,(XtCallbackProc)switch_button_sram,NULL);
  XmStringFree(string);
  
  string=XmStringGenerate(_("Close"),XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  temp=XtVaCreateManagedWidget("Close",xmPushButtonGadgetClass,
			       sram_window,
			       XmNleftAttachment,XmATTACH_POSITION,
			       XmNleftPosition,3,
			       XmNrightAttachment,XmATTACH_POSITION,
			       XmNrightPosition,4,XmNlabelString,string,
			       XmNbottomAttachment,XmATTACH_FORM,
			       XmNbottomOffset,6,NULL);
  
  XtAddCallback(temp,
		XmNactivateCallback,(XtCallbackProc)close_button_sram,NULL);
  XmStringFree(string);
  XtVaSetValues(frame,XmNbottomOffset,6,XmNbottomAttachment,XmATTACH_WIDGET,
		XmNbottomWidget,temp,NULL);
  XtAddCallback(shell,XmNpopupCallback,(XtCallbackProc)set_size_sram_window,
		NULL);
  XtAddEventHandler(shell,StructureNotifyMask,False,resize_sram,NULL);
}

void create_sram_line(int n,int address,Widget *widgetlist,int *n_wid)
{
  int i,n_args,num_wid,t;
  char buf[6];
  Arg args[20];

  num_wid=*n_wid;
  n_args=0;
  sprintf(buf,"%X",address+controllers[type_micro].size_gpr+
	  controllers[type_micro].size_io_reg);
  fill_zero(buf,4);
  
  /* main resource */
  XtSetArg(args[n_args],XmNcursorPositionVisible,False);n_args++;
  XtSetArg(args[n_args],XmNshadowThickness,1);n_args++;
  t=n_args;

  XtSetArg(args[n_args],XmNcolumns,4);n_args++;
  XtSetArg(args[n_args],XmNmaxLength,4);n_args++;
  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_FORM);n_args++;
  XtSetArg(args[n_args],XmNvalue,buf);n_args++;
  XtSetArg(args[n_args],XmNeditable,False);n_args++;
  if(n==0)
    {
      XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_FORM);n_args++;
      XtSetArg(args[n_args],XmNtopOffset,1);n_args++;

    }
  else
    {
      XtSetArg(args[n_args],XmNtopOffset,2);n_args++;
      XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_WIDGET);n_args++;
      XtSetArg(args[n_args],XmNtopWidget,label_sram[n-1]);n_args++;
    }
  
  widgetlist[num_wid++]=label_sram[n]=
    XtCreateManagedWidget("memory_label",xmTextFieldWidgetClass,dataform_sram,
			  args,n_args);

  n_args=t;
  XtSetArg(args[n_args],XmNleftOffset,0);n_args++;
  XtSetArg(args[n_args],XmNcolumns,3);n_args++;
  XtSetArg(args[n_args],XmNmaxLength,2);n_args++;
  XtSetArg(args[n_args],XmNvalue,"0");n_args++;
  t=n_args;
  for(i=0;i<16;i++)
    {
      n_args=t;
      if(i==0) 
	{
	  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_WIDGET);n_args++;
	  XtSetArg(args[n_args],XmNleftWidget,label_sram[n]);n_args++;
	}
      else
	{
	  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_WIDGET);n_args++;
	  XtSetArg(args[n_args],XmNleftWidget,field_sram[n*16+i-1]);n_args++;
	}
      
      if(n==0) 
	{XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_FORM);n_args++;}
      else 
	{
	  XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_WIDGET);n_args++;
	  XtSetArg(args[n_args],XmNtopWidget,field_sram[(n-1)*16+i]);n_args++;
	}
      widgetlist[num_wid++]=field_sram[n*16+i]=
	XtCreateManagedWidget("memory_field",xmTextFieldWidgetClass,
			      dataform_sram,args,n_args);
      xs_widget_tip(field_sram[n*16+i],sram_tip_str[n]);
      XtAddCallback(field_sram[n*16+i],XmNfocusCallback,(XtCallbackProc)
		    get_focus_field_sram_callback,NULL);
      XtAddCallback(field_sram[n*16+i],XmNlosingFocusCallback,
		    (XtCallbackProc)lose_focus_field_sram_callback,NULL);
      
    }
  XtVaSetValues(field_sram[n*16+i-1],XmNrightAttachment,
		XmATTACH_FORM,NULL);
  *n_wid=num_wid;
}

void resize_sram(Widget w,char *m,XConfigureEvent *event,
		 Boolean *continueDispatch)
{
  int n,n_wid;
  Widget *widgetlist;
  
  if(event->type==ConfigureNotify)
    if(event->height!=height_min_sram_window+nlines_sram*sram_window_step) 
      {
	nlines_sram_old=nlines_sram;
	nlines_sram=(event->height-height_min_sram_window)/sram_window_step;
	
	if(nlines_sram>controllers[type_micro].size_sram/16)
	  {
	    nlines_sram=controllers[type_micro].size_sram/16;
	    if(nlines_sram==nlines_sram_old) return;
	  }
	
	if(nlines_sram>nlines_sram_old)
	  {  /* increase io_lines */
	    
	    if((widgetlist=malloc(nlines_sram*17*sizeof(Widget)))==NULL)
	      {
		puts("resize_sram(): can't allocate memory for 'widgetlist'");
		exit(1);
	      }
	    n_wid=0;
	    for(n=nlines_sram_old;n<nlines_sram;n++)
	      {
		
		if(label_sram[n]==NULL)
		  create_sram_line(n,0,widgetlist,&n_wid);
	      }
	    if(n_wid) XtManageChildren(widgetlist,n_wid);
	    free(widgetlist);
	  }
	
	XtVaSetValues(vscr_sram_bar,XmNmaximum,
		      controllers[type_micro].size_sram/16,XmNsliderSize,
		      nlines_sram,NULL);
	
	if(address_sram+nlines_sram>controllers[type_micro].size_sram/16)
	  {
	    address_sram=controllers[type_micro].size_sram-nlines_sram*16;
	    XtVaSetValues(vscr_sram_bar,XmNvalue,address_sram/16,NULL);
	  }
	full_update_sram_window();
      }

}

void scroll_bar_sram_callback (Widget w,XtPointer client_data,
			       XtPointer call_data)
{ 
  XtVaGetValues(vscr_sram_bar,XmNvalue,&address_sram,NULL);
  address_sram*=16;
  full_update_sram_window();
}

void full_update_sram_window(void)
{
  int n,address;
  char buf[10];
  unsigned char value;
  
  XtVaGetValues(menu[MEMORY_MENU].sub_menu[SRAM_BUTTON].button,
		XmNset,&value,NULL);
  if(value!=XmSET) return;
  
  if(sram_pointer==NULL) return;

  /* all update */
  memset(&mem_file_changed[(controllers[type_micro].size_gpr+controllers[type_micro].size_io_reg)/8],0,controllers[type_micro].size_sram/8);

  for(n=0;n<nlines_sram*16;n++)
    {
      if(type_sram_window==HEX)
	sprintf(buf,"%X",sram_pointer[n+address_sram]);
      else sprintf(buf,"%u",sram_pointer[n+address_sram]);
      XmTextFieldSetString(field_sram[n],buf);
      delete_tip(field_sram[n],sram_tip_str[n]);
      xs_widget_tip(field_sram[n],
		    conv_to_bin((int)sram_pointer[n+address_sram],
				sram_tip_str[n]));
    }

  for(n=0,address=address_sram;n<nlines_sram;n++,address+=16)
    {
      sprintf(buf,"%X",address+controllers[type_micro].size_gpr+
	      controllers[type_micro].size_io_reg);
      fill_zero(buf,4);
      XmTextFieldSetString(label_sram[n],buf);
    }
}

void update_sram_window(void)
{
  int n,address;
  char buf[10];
  unsigned char value;
    
  XtVaGetValues(menu[MEMORY_MENU].sub_menu[SRAM_BUTTON].button,
		XmNset,&value,NULL);
  if(value!=XmSET) return;

  if(sram_pointer==NULL) return;
  address=
    (controllers[type_micro].size_gpr+controllers[type_micro].size_io_reg)/8;
  
  for(n=0;n<nlines_sram*16;n++)
    {
      if(mem_file_changed[address+(n+address_sram)/8]&
	 (1<<(n+address_sram)%8))
	mem_file_changed[address+(n+address_sram)/8]&=~(1<<(n+address_sram)%8);
      else
	continue;

      if(type_sram_window==HEX)
	sprintf(buf,"%X",sram_pointer[n+address_sram]);
      else sprintf(buf,"%u",sram_pointer[n+address_sram]);
      XtVaSetValues(field_sram[n],XmNvalue,buf,NULL);
      delete_tip(field_sram[n],sram_tip_str[n]);
      xs_widget_tip(field_sram[n],
		    conv_to_bin((int)sram_pointer[n+address_sram],
				sram_tip_str[n]));
    }

}


void get_focus_field_sram_callback (Widget w,XtPointer client_data,
				    XtPointer call_data)
{
  XtVaSetValues(w,XmNcursorPositionVisible,True,NULL);
  XtAddCallback(w,XmNvalueChangedCallback,
		(XtCallbackProc)field_sram_changed_callback,NULL);
}


void lose_focus_field_sram_callback (Widget w,XtPointer client_data,
			       XtPointer call_data)
{
  XtVaSetValues(w,XmNcursorPositionVisible,False,NULL);
  XtRemoveCallback(w,XmNvalueChangedCallback,(XtCallbackProc)
		   field_sram_changed_callback,NULL);
}


void field_sram_changed_callback(Widget w,XtPointer client_data,
				 XtPointer call_data)
{
  char *buf,buf2[7];
  int n;
  
  buf=XmTextFieldGetString(w);
  n=0;
  while(n<nlines_sram*16&&w!=field_sram[n]) n++;
  if(type_sram_window==DEC)
    sram_pointer[n+address_sram]=strtoul(buf,NULL,10);
  else
    {
      sprintf(buf2,"0x%s",buf);
      sram_pointer[n+address_sram]=strtoul(buf2,NULL,16);
      delete_tip(field_sram[n],sram_tip_str[n]);
      xs_widget_tip(field_sram[n],
		    conv_to_bin((int)sram_pointer[n+address_sram],
				sram_tip_str[n]));
    }
  XtFree(buf);
}

void switch_button_sram(Widget w,XtPointer client_data,XtPointer call_data)
{
  XmString string;
  int n;

  if(type_sram_window==HEX) 
    {
      type_sram_window=DEC;
      string=XmStringGenerate(_("   Hex   "),XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(w,XmNlabelString,string,NULL);
      XmStringFree(string);
      for(n=0;n<nlines_sram*16;n++)
	XtVaSetValues(field_sram[n],XmNmaxLength,3,NULL);
    }
  else
    {
      type_sram_window=HEX;
      string=XmStringGenerate(_(" Decimal "),XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(w,XmNlabelString,string,NULL);
      XmStringFree(string);
      for(n=0;n<nlines_sram*16;n++)
	XtVaSetValues(field_sram[n],XmNmaxLength,2,NULL);
    }
  full_update_sram_window();
}

void close_button_sram(Widget w,XtPointer client_data,XtPointer call_data)
{
  XtVaSetValues(menu[MEMORY_MENU].sub_menu[SRAM_BUTTON].button,XmNset,
		XmUNSET,NULL);
  XtUnmanageChild(sram_window);
  return;
}

void set_size_sram_window(Widget w,XtPointer client_data,XtPointer call_data)
{
  Dimension width,height,height_label;

  XtRemoveCallback(w,XmNpopupCallback,(XtCallbackProc)set_size_sram_window,
		   NULL);
  XtVaGetValues(w,XmNwidth,&width,XmNheight,&height,NULL);
  XtVaGetValues(label_sram[0],XmNheight,&height_label,NULL);
  sram_window_step=height_label+2;
  height_min_sram_window=height-nlines_sram*sram_window_step;

  XtVaSetValues(w,XmNmaxWidth,width,XmNminWidth,width,XmNminHeight,
		height+(4-nlines_sram)*sram_window_step,
		XmNheightInc,sram_window_step,XmNmaxHeight,
		height_min_sram_window+MAX_SRAM_LINES*sram_window_step,NULL);
}

/* I/O window */
void create_io_ports_window(void)
{
  Widget scroll,temp,shell,frame,io_form,*widgetlist;
  XmString string;
  int n,num_ports,n_wid;

  address_io_old=address_io;
  for(n=0;controllers[type_micro].io_registers[n].code!=-1;n++);
  num_ports=n;

  shell=XtVaCreateWidget("IO_WINDOW",xmDialogShellWidgetClass,toplevel,
			 XmNtitle,_("I/O Registers"),
			 XmNkeyboardFocusPolicy,XmPOINTER,NULL);
  io_window=XtVaCreateWidget("form",xmFormWidgetClass,shell,
			     XmNautoUnmanage,False,
			     XmNdialogStyle,XmDIALOG_MODELESS,
			     XmNfractionBase,3,NULL);

  frame=XtVaCreateManagedWidget("frame",xmFrameWidgetClass,io_window,
				XmNtopAttachment,XmATTACH_FORM,
				XmNtopOffset,5,
				XmNleftAttachment,XmATTACH_FORM,
				XmNleftOffset,5,
				XmNrightAttachment,XmATTACH_FORM,
				XmNrightOffset,5,NULL);
  io_form=XtVaCreateManagedWidget("form",xmFormWidgetClass,frame,NULL);
  
  string=XmStringGenerate(_("Name  Addr  Bit values   Hex"),
			  XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  temp=XtVaCreateManagedWidget("label",xmLabelGadgetClass,io_form,
			       XmNlabelString,string,XmNtopOffset,5,
			       XmNleftAttachment,XmATTACH_FORM,
			       XmNleftOffset,5,NULL);
  XmStringFree(string);
  
  temp=XtVaCreateManagedWidget("separator",xmSeparatorGadgetClass,io_form,
			       XmNtopAttachment,XmATTACH_WIDGET,
			       XmNtopWidget,temp,
			       XmNleftAttachment,XmATTACH_FORM,
			       XmNrightAttachment,XmATTACH_FORM,NULL);
  /* create scroll window */
  scroll=XtVaCreateManagedWidget("Scroll",xmScrolledWindowWidgetClass,
				 io_form,
				 XmNscrollingPolicy,XmAPPLICATION_DEFINED,
				 XmNscrollBarDisplayPolicy,XmSTATIC,
				 XmNvisualPolicy,XmVARIABLE,
				 XmNleftAttachment,XmATTACH_FORM,
				 XmNleftOffset,3,
				 XmNrightAttachment,XmATTACH_FORM,
				 XmNrightOffset,2,
				 XmNtopAttachment,XmATTACH_WIDGET,
				 XmNtopWidget,temp,
				 XmNtopOffset,2,
				 XmNbottomAttachment,XmATTACH_FORM,
				 NULL);
  
  dataform_io=XtVaCreateManagedWidget("form area",xmFormWidgetClass,scroll,
				      NULL);
  
  vscr_io_bar=XtVaCreateManagedWidget("scr_bar",xmScrollBarWidgetClass,
				      scroll,XmNorientation,XmVERTICAL,
				      XmNincrement,1,XmNminimum,0,
				      XmNvalue,0,
				      XmNmaximum,num_ports,
				      XmNsliderSize,nlines_io,
				      NULL);
  XtAddCallback(vscr_io_bar,XmNvalueChangedCallback,
		(XtCallbackProc)scroll_bar_io_callback,NULL);
  XtAddCallback(vscr_io_bar,XmNdragCallback,
		(XtCallbackProc)scroll_bar_io_callback,NULL);
  XmScrolledWindowSetAreas(scroll,NULL,vscr_io_bar,dataform_io);
 
  /* allocate memory for labels and bit buttons */
  if(nlines_io>num_ports) nlines_io=num_ports;
  if((label_io=malloc(MAX_IO_PORT_LINES*sizeof(Widget)))==NULL)
    {
      puts("create_io_ports_window(): can't allocate memory for 'label_io'");
      exit(1);
    }
  memset(label_io,0,MAX_IO_PORT_LINES*sizeof(Widget));

  if((bit_button=malloc(MAX_IO_PORT_LINES*sizeof(Widget *)))==NULL)
    {
      puts("create_io_ports_window(): \
can't allocate memory for 'bit_button'");
      exit(1);
    }
  for(n=0;n<MAX_IO_PORT_LINES;n++)
    if((bit_button[n]=malloc(8*sizeof(Widget)))==NULL)
      {
      printf("create_io_ports_window(): \
can't allocate memory for 'bit_button[%d]'",n);
      exit(1);
      }
  if((label_io_val=malloc(MAX_IO_PORT_LINES*sizeof(Widget)))==NULL)
    {
      puts("create_io_ports_window(): \
can't allocate memory for 'label_io_val'");
      exit(1);
    }
  if((label_io_addr=malloc(MAX_IO_PORT_LINES*sizeof(Widget)))==NULL)
    {
      puts("create_io_ports_window(): \
can't allocate memory for 'label_io_addr'");
      exit(1);
    }

  /* create name labels of ports and bit buttons */
  if((widgetlist=malloc(nlines_io*11*sizeof(Widget)))==NULL)
    {
      puts("create_io_ports_window(): \
can't allocate memory for 'widgetlist'");
      exit(1);
    }
  n_wid=0;
  for(n=0;n<nlines_io;n++) create_io_ports_line(n,widgetlist,&n_wid);
  XtManageChildren(widgetlist,n_wid);
  free(widgetlist);

  string=XmStringGenerate(_("Close"),XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  temp=XtVaCreateManagedWidget("Close",xmPushButtonGadgetClass,
			       io_window,
			       XmNrightAttachment,XmATTACH_POSITION,
			       XmNrightPosition,2,XmNlabelString,string,
			       XmNleftAttachment,XmATTACH_POSITION,
			       XmNleftPosition,1,
			       XmNbottomAttachment,XmATTACH_FORM,
			       XmNbottomOffset,6,NULL);
  
  XtAddCallback(temp,
		XmNactivateCallback,(XtCallbackProc)close_button_io,NULL);
  XmStringFree(string);
  XtVaSetValues(frame,XmNbottomAttachment,XmATTACH_WIDGET,
		XmNbottomWidget,temp,XmNbottomOffset,6,NULL);
  XtAddCallback(shell,XmNpopupCallback,(XtCallbackProc)set_size_io_window,
		NULL);
  XtAddEventHandler(shell,StructureNotifyMask,False,resize_io,NULL);
}


void create_io_ports_line(int n,Widget *widgetlist,int *num_wid)
{
  int i,n_wid,n_args;
  char buf[10];
  XmString string;
  Arg args[10];
  
  n_args=0;n_wid=*num_wid;
  /* create name labels */
  string=XmStringGenerate(descr_io_ports[controllers[type_micro].io_registers[n].code].name,
			  XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  
  XtSetArg(args[n_args],XmNlabelString,string);n_args++;
  XtSetArg(args[n_args],XmNtopOffset,2);n_args++;
  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_FORM);n_args++;
  
  if(n==0) 
    { XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_FORM);n_args++;}
  else
    {
      XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_WIDGET);n_args++;
      XtSetArg(args[n_args],XmNtopWidget,label_io[n-1]);n_args++;
    }
  
  widgetlist[n_wid++]=
    label_io[n]=XtCreateWidget("label",xmLabelWidgetClass,
			       dataform_io,args,n_args);
  XmStringFree(string);
  xs_widget_tip(label_io[n],descr_io_ports[controllers[type_micro].io_registers[n+address_io].code].fullname);
  
  /* create address labels */
  sprintf(buf,"%Xh",controllers[type_micro].io_registers[n].address);
  fill_zero(buf,3);
  string=XmStringGenerate(buf,
			  XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  n_args=0;
  XtSetArg(args[n_args],XmNlabelString,string);n_args++;
  XtSetArg(args[n_args],XmNtopOffset,2);n_args++;
  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_WIDGET);n_args++;
  XtSetArg(args[n_args],XmNleftWidget,label_io[n]);n_args++;

  if(n==0)
    {
      XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_FORM);n_args++;
    }
  else
    {
      XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_WIDGET);n_args++;
      XtSetArg(args[n_args],XmNtopWidget,label_io_addr[n-1]);n_args++;
      /*  if(n==nlines_io) */
/*  	{ */
/*  	  XtVaSetValues(label_io[n],XmNbottomAttachment,XmATTACH_FORM, */
/*  			NULL); */
/*  	  XtVaSetValues(label_io_addr[n],XmNbottomAttachment,XmATTACH_FORM, */
/*  			NULL); */
/*  	} */
    }

  widgetlist[n_wid++]=
    label_io_addr[n]=XtCreateWidget("label",xmLabelGadgetClass,dataform_io,
				    args,n_args);
  XmStringFree(string);
  
  /* create bit button */
  n_args=0;
  XtSetArg(args[n_args],XmNtopOffset,0);n_args++;
  XtSetArg(args[n_args],XmNhighlightOnEnter,True);n_args++;
  XtSetArg(args[n_args],XmNshadowThickness,0);n_args++;

  for(i=0;i<8;i++)
    {
      n_args=4;
      if(i==0)
	{
	  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_WIDGET);n_args++;
	  XtSetArg(args[n_args],XmNleftWidget,label_io_addr[n]);n_args++;
	  XtSetArg(args[n_args],XmNleftOffset,5);n_args++;
	}
      else
	{
	  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_WIDGET);n_args++;
	  XtSetArg(args[n_args],XmNleftWidget,bit_button[n][i-1]);n_args++;
	}
      if(n==0)
	{XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_FORM);n_args++;}
      else
	{
	  XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_WIDGET);n_args++;
	  XtSetArg(args[n_args],XmNtopWidget,bit_button[n-1][i]);n_args++;
	}

      widgetlist[n_wid++]=
	bit_button[n][i]=XtCreateWidget("0",xmPushButtonGadgetClass,
					  dataform_io,args,n_args);
      
      if(!((controllers[type_micro].io_registers[n].read_mask|
	    controllers[type_micro].io_registers[n].write_mask)
	   &(1<<(7-i))))
	XtSetSensitive(bit_button[n][i],False);
      
      XtAddCallback(bit_button[n][i],
		    XmNactivateCallback,
		    (XtCallbackProc)bit_button_callback,NULL);
    }

  /* create labels of values */
  string=XmStringGenerate("00h",XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  n_args=0;
  XtSetArg(args[n_args],XmNlabelString,string);n_args++;
  XtSetArg(args[n_args],XmNtopOffset,2);n_args++;
  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_WIDGET);n_args++;
  XtSetArg(args[n_args],XmNleftWidget,bit_button[n][7]);n_args++;
  XtSetArg(args[n_args],XmNleftOffset,10);n_args++;

 
  if(n==0)
    {
      XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_FORM);n_args++;
      XtSetArg(args[n_args],XmNtopOffset,1);n_args++;
    }
  else
    {
      XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_WIDGET);n_args++;
      XtSetArg(args[n_args],XmNtopWidget,label_io_val[n-1]);n_args++;
    }
  if(n==nlines_io)
    {XtSetArg(args[n_args],XmNbottomAttachment,XmATTACH_FORM);n_args++;}
  
  widgetlist[n_wid++]=
    label_io_val[n]=XtCreateWidget("label",xmLabelGadgetClass,dataform_io,
				   args,n_args);
  XmStringFree(string);
  
  *num_wid=n_wid;
}

void resize_io(Widget w,char *m,XConfigureEvent *event,
	       Boolean *continueDispatch)
{
  int old,num_ports,n,n_wid;
  Widget *widgetlist;
  
  if(event->type==ConfigureNotify)
    if(event->height!=height_min_io_window+nlines_io*io_window_step)
      {
	for(n=0;controllers[type_micro].io_registers[n].code!=-1;n++);
	num_ports=n;
	old=nlines_io;
	nlines_io=(event->height-height_min_io_window)/io_window_step;
	
	if(nlines_io>num_ports)
	  {
	    nlines_io=num_ports;
	    if(nlines_io==old) return;
	  }
	
	if(nlines_io>old)
	  { /* increase io_lines */

	    if((widgetlist=malloc(nlines_io*11*sizeof(Widget)))==NULL)
	      {
		puts("resize_io(): can't allocate memory for 'widgetlist'");
		exit(1);
	      }
	    n_wid=0;
	    for(n=old;n<nlines_io;n++)
	      {
		
		if(label_io[n]==NULL)
		  create_io_ports_line(n,widgetlist,&n_wid);
	      }
	    if(n_wid) XtManageChildren(widgetlist,n_wid);
	    free(widgetlist);
	  }
	
	XtVaSetValues(vscr_io_bar,XmNmaximum,num_ports,XmNsliderSize,
		      nlines_io,NULL);
	
	if(address_io+nlines_io>num_ports)
	  {
	    address_io_old=address_io;
	    address_io=num_ports-nlines_io;
	    XtVaSetValues(vscr_io_bar,XmNvalue,num_ports-nlines_io,NULL);
	  }
	
	full_update_io_window();
      }
}

void update_io_window(void)
{
  int i,n,address;
  XmString string;
  char buf[10];
  unsigned char value;
    
  XtVaGetValues(menu[MEMORY_MENU].sub_menu[IO_BUTTON].button,
		XmNset,&value,NULL);
  if(value!=XmSET) return;
  
  for(n=0;n<nlines_io;n++)
    {
      address=controllers[type_micro].io_registers[address_io+n].address;
   
      if(mem_file_changed[(controllers[type_micro].size_gpr+address)/8]&
	 (1<<(controllers[type_micro].size_gpr+address)%8))
	mem_file_changed[(controllers[type_micro].size_gpr+address)/8]&=
	  ~(1<<(controllers[type_micro].size_gpr+address)%8);
      else
	{
	  if(address!=0x3F)
	    continue;
	}
      
      for(i=0;i<8;i++)
	{
	  if(io_pointer[address]&(1<<(7-i)))
	    string=XmStringGenerate("1",XmFONTLIST_DEFAULT_TAG,
				    XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
	  else
	    string=XmStringGenerate("0",XmFONTLIST_DEFAULT_TAG,
				    XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
	  XtVaSetValues(bit_button[n][i],XmNlabelString,string,NULL);
	  XmStringFree(string);
	}
      
      sprintf(buf,"%X",io_pointer[address]);
      fill_zero(buf,2);
      buf[2]='h';buf[3]=0;
      string=XmStringGenerate(buf,XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(label_io_val[n],XmNlabelString,string,NULL);
      XmStringFree(string);
    }
}

void full_update_io_window(void)
{
  int i,n;
  XmString string;
  char buf[10];
  unsigned char value;
    
  XtVaGetValues(menu[MEMORY_MENU].sub_menu[IO_BUTTON].button,
		XmNset,&value,NULL);
  if(value!=XmSET) return;
  
  /* all update */
  memset(&mem_file_changed[controllers[type_micro].size_gpr/8],0,
	 controllers[type_micro].size_io_reg/8);

  for(n=0;n<nlines_io;n++)
    {
      string=XmStringGenerate(descr_io_ports[controllers[type_micro].io_registers[n+address_io].code].name,
			      XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(label_io[n],XmNlabelString,string,NULL);
      XmStringFree(string);
      
      delete_tip(label_io[n],descr_io_ports[controllers[type_micro_old].io_registers[n+address_io_old].code].fullname);
      xs_widget_tip(label_io[n],descr_io_ports[controllers[type_micro].io_registers[n+address_io].code].fullname);

      sprintf(buf,"%Xh",
	      controllers[type_micro].io_registers[n+address_io].address);
      fill_zero(buf,3);
      string=XmStringGenerate(buf,XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(label_io_addr[n],XmNlabelString,string,NULL);
      XmStringFree(string);
      for(i=0;i<8;i++)
	{
	  if((controllers[type_micro].io_registers[address_io+n].read_mask|
		controllers[type_micro].io_registers[address_io+n].write_mask)
	       &(1<<(7-i)))
	    XtSetSensitive(bit_button[n][i],True);
	  else 
	    XtSetSensitive(bit_button[n][i],False);
	  
	  if(io_pointer[controllers[type_micro].io_registers[address_io+n].address]&(1<<(7-i)))
	    string=XmStringGenerate("1",XmFONTLIST_DEFAULT_TAG,
				    XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
	  else
	    string=XmStringGenerate("0",XmFONTLIST_DEFAULT_TAG,
				    XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
	  XtVaSetValues(bit_button[n][i],XmNlabelString,string,NULL);
	  XmStringFree(string);
	}
      
      sprintf(buf,"%X",io_pointer[controllers[type_micro].io_registers[address_io+n].address]);
      fill_zero(buf,2);
      buf[2]='h';buf[3]=0;
      string=XmStringGenerate(buf,XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(label_io_val[n],XmNlabelString,string,NULL);
      XmStringFree(string);
    }
  address_io_old=address_io;
  type_micro_old=type_micro;
}

void scroll_bar_io_callback(Widget w,XtPointer client_data,
			    XtPointer call_data)
{
  address_io=((XmScrollBarCallbackStruct *)call_data)->value;
  full_update_io_window();
}

void bit_button_callback(Widget w,XtPointer client_data,XtPointer call_data)
{
  int n,m;
  char buf[5];
  XmString string;

  n=m=0;
  while(w!=bit_button[n][m])
    if(++m==8) {m=0;n++;}
  if(io_pointer[controllers[type_micro].io_registers[address_io+n].address]
     &(1<<(7-m)))
    {
      io_pointer[controllers[type_micro].io_registers[address_io+n].address]=
	io_pointer[controllers[type_micro].io_registers[address_io+n].address]
	&~(1<<(7-m));
      string=XmStringGenerate("0",XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(bit_button[n][m],XmNlabelString,string,NULL);
      XmStringFree(string);
    }
  else
    {
      io_pointer[controllers[type_micro].io_registers[address_io+n].address]=
	io_pointer[controllers[type_micro].io_registers[address_io+n].address]
	|(1<<(7-m));
      string=XmStringGenerate("1",XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(bit_button[n][m],XmNlabelString,string,NULL);
      XmStringFree(string);
    }
  sprintf(buf,"%X",io_pointer[controllers[type_micro].io_registers[address_io+n].address]);
  fill_zero(buf,2);
  buf[2]='h';buf[3]=0;
  string=XmStringGenerate(buf,XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  XtVaSetValues(label_io_val[n],XmNlabelString,string,NULL);
  XmStringFree(string);
  if(controllers[type_micro].io_registers[address_io+n].code==SREG) 
    update_main();
}

void set_size_io_window(Widget w,XtPointer client_data,XtPointer call_data)
{
  Dimension width,height,height_label;

  XtRemoveCallback(w,XmNpopupCallback,(XtCallbackProc)set_size_io_window,
		   NULL);

  XtVaGetValues(label_io[0],XmNheight,&height_label,NULL);
  io_window_step=height_label+2;
  
  XtVaGetValues(w,XmNwidth,&width,XmNheight,&height,NULL);
  height_min_io_window=height-nlines_io*io_window_step;
  XtVaSetValues(w,XmNmaxWidth,width,XmNminWidth,width,XmNminHeight,
		height+(4-nlines_io)*io_window_step,
		XmNheightInc,io_window_step,XmNwidth,width,
		XmNmaxHeight,
		height_min_io_window+MAX_IO_PORT_LINES*io_window_step,NULL);
}

void close_button_io(Widget w,XtPointer client_data,XtPointer call_data)
{
  XtVaSetValues(menu[MEMORY_MENU].sub_menu[IO_BUTTON].button,XmNset,
		XmUNSET,NULL);
  XtUnmanageChild(io_window);
  return;
}

/* EEPROM window */
void create_eeprom_window(void)
{
  Widget scroll,temp,shell,eeprom_form,frame,*widgetlist;
  XmString string;
  char buf[6];
  int n,address,n_wid,n_args,t;
  Arg args[20];

  if((field_eeprom=malloc(16*MAX_EEPROM_LINES*sizeof(Widget)))==NULL)
    {
      puts("create_eeprom_window(): \
can't allocate memory for 'field_eeprom'");
      exit(1);
    }
  if((label_eeprom=malloc(MAX_EEPROM_LINES*sizeof(Widget)))==NULL)
    {
      puts("create_eeprom_window(): \
can't allocate memory for 'label_eeprom'");
      exit(1);
    }
  if((eeprom_tip_str=malloc(MAX_EEPROM_LINES*16*sizeof(char *)))==NULL)
    {
      puts("create_eeprom_window(): \
can't allocate memory for 'eeprom_tip_str'");
      exit(1);
    }
  
  for(n=0;n<MAX_EEPROM_LINES*16;n++)
    {
      if((eeprom_tip_str[n]=malloc(10*sizeof(char)))==NULL)
	{
	  printf("create_eeprom_window(): \
can't allocate memory for 'eeprom_tip_str[%d]'",n);
	  exit(1);
	}
      eeprom_tip_str[n][0]=0;
    }


  memset(label_eeprom,0,MAX_EEPROM_LINES*sizeof(Widget));
  
  shell=XtVaCreateWidget("Shell",xmDialogShellWidgetClass,toplevel,
			 XmNtitle,"EEPROM",NULL);
  eeprom_window=XtVaCreateWidget("form",xmFormWidgetClass,shell,
				 XmNautoUnmanage,False,
				 XmNdialogStyle,XmDIALOG_MODELESS,
				 XmNfractionBase,5,NULL);
  
  frame=XtVaCreateManagedWidget("frame",xmFrameWidgetClass,eeprom_window,
				XmNtopAttachment,XmATTACH_FORM,
				XmNtopOffset,5,
				XmNleftAttachment,XmATTACH_FORM,
				XmNleftOffset,5,
				XmNrightAttachment,XmATTACH_FORM,
				XmNrightOffset,5,NULL);
  eeprom_form=XtVaCreateManagedWidget("form",xmFormWidgetClass,frame,NULL);

/* create firs row, labels */
  n_args=0;
  XtSetArg(args[n_args],XmNcursorPositionVisible,False);n_args++;
  XtSetArg(args[n_args],XmNeditable,False);n_args++;
  XtSetArg(args[n_args],XmNtopOffset,8);n_args++;
  XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_FORM);n_args++;

  t=n_args;

  XtSetArg(args[n_args],XmNleftOffset,3);n_args++;
  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_FORM);n_args++;
  XtSetArg(args[n_args],XmNvalue,"");n_args++;
  XtSetArg(args[n_args],XmNcolumns,4);n_args++;
  XtSetArg(args[n_args],XmNmaxLength,4);n_args++;

  temp=XtCreateManagedWidget("memory_label",xmTextFieldWidgetClass,
			     eeprom_form,args,n_args);

  n_args=t;
  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_WIDGET);n_args++;
  XtSetArg(args[n_args],XmNcolumns,3);n_args++;
  XtSetArg(args[n_args],XmNmaxLength,3);n_args++;
  XtSetArg(args[n_args],XmNleftOffset,2);n_args++;
  t=n_args;
  for(n=0;n<16;n++)
    {
      n_args=t;
      sprintf(buf,"%X",n);
      fill_zero(buf,2);
      XtSetArg(args[n_args],XmNvalue,buf);n_args++;
      XtSetArg(args[n_args],XmNleftWidget,temp);n_args++;
      temp=XtCreateManagedWidget("r",xmTextFieldWidgetClass,eeprom_form,args,
				 n_args);
    }

  /* create scroll window */
  scroll=XtVaCreateManagedWidget("Scroll",xmScrolledWindowWidgetClass,
				 eeprom_form,
				 XmNscrollingPolicy,XmAPPLICATION_DEFINED,
				 XmNscrollBarDisplayPolicy,XmSTATIC,
				 XmNleftAttachment,XmATTACH_FORM,
				 XmNleftOffset,3,
				 XmNrightAttachment,XmATTACH_FORM,
				 XmNrightOffset,2,
				 XmNtopAttachment,XmATTACH_WIDGET,
				 XmNtopWidget,temp,
				 XmNbottomAttachment,XmATTACH_FORM,
				 XmNbottomOffset,3,
				 NULL);
  
  dataform_eeprom=
    XtVaCreateManagedWidget("form area",xmFormWidgetClass,scroll,NULL);
  
  vscr_eeprom_bar=XtVaCreateManagedWidget("scr_bar",xmScrollBarWidgetClass,
					  scroll,XmNorientation,XmVERTICAL,
					  XmNincrement,1,XmNminimum,0,
					  XmNvalue,0,NULL);
  XtAddCallback(vscr_eeprom_bar,XmNvalueChangedCallback,
		(XtCallbackProc)scroll_bar_eeprom_callback,NULL);
  XtAddCallback(vscr_eeprom_bar,XmNdragCallback,
		(XtCallbackProc)scroll_bar_eeprom_callback,NULL);
  XmScrolledWindowSetAreas(scroll,NULL,vscr_eeprom_bar,dataform_eeprom);
  
  /* size of 'vscr_eeprom_bar' must be equal size of end space */
  XtVaGetValues(vscr_eeprom_bar,XmNwidth,(Dimension *)&n,NULL);
  XtVaSetValues(temp,XmNrightAttachment,XmATTACH_FORM,
		XmNrightOffset,n,NULL);
  
  /* create input fields */
  if((widgetlist=malloc(nlines_eeprom*17*sizeof(Widget)))==NULL)
    {
      puts("create_eeprom_window(): \
can't allocate memory for 'widgetlist'");
      exit(1);
    }
  n_wid=0;
  for(n=0,address=address_eeprom;n<nlines_eeprom;n++,address+=16)
    create_eeprom_line(n,address,widgetlist,&n_wid);
  XtManageChildren(widgetlist,n_wid);
  free(widgetlist);
  
  string=XmStringGenerate(_("Decimal"),XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);  
  temp=XtVaCreateManagedWidget("Dec",xmPushButtonGadgetClass,
			       eeprom_window,
			       XmNleftAttachment,XmATTACH_POSITION,
			       XmNleftPosition,1,XmNlabelString,string,
			       XmNrightAttachment,XmATTACH_POSITION,
			       XmNrightPosition,2,
			       XmNbottomAttachment,XmATTACH_FORM,
			       XmNbottomOffset,6,NULL);
  
  XtAddCallback(temp,
		XmNactivateCallback,(XtCallbackProc)switch_button_eeprom,NULL);
  XmStringFree(string);
  
  string=XmStringGenerate(_("Close"),XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  temp=XtVaCreateManagedWidget("Close",xmPushButtonGadgetClass,
			       eeprom_window,
			       XmNleftAttachment,XmATTACH_POSITION,
			       XmNleftPosition,3,
			       XmNrightAttachment,XmATTACH_POSITION,
			       XmNrightPosition,4,XmNlabelString,string,
			       XmNbottomAttachment,XmATTACH_FORM,
			       XmNbottomOffset,6,NULL);
  
  XtAddCallback(temp,
		XmNactivateCallback,(XtCallbackProc)close_button_eeprom,NULL);
  XmStringFree(string);
  XtVaSetValues(frame,XmNbottomOffset,6,XmNbottomAttachment,XmATTACH_WIDGET,
		XmNbottomWidget,temp,NULL);
  XtAddCallback(shell,XmNpopupCallback,(XtCallbackProc)set_size_eeprom_window,
		NULL);
  XtAddEventHandler(shell,StructureNotifyMask,False,resize_eeprom,NULL);
}

void create_eeprom_line(int n,int address,Widget *widgetlist,int *n_wid)
{
  int i,n_args,num_wid,t;
  char buf[6];
  Arg args[20];

  num_wid=*n_wid;
  n_args=0;
  sprintf(buf,"%X",address);
  fill_zero(buf,4);

  /* main resource */

  XtSetArg(args[n_args],XmNcursorPositionVisible,False);n_args++;
  XtSetArg(args[n_args],XmNshadowThickness,1);n_args++;
  t=n_args;

  XtSetArg(args[n_args],XmNcolumns,4);n_args++;
  XtSetArg(args[n_args],XmNmaxLength,4);n_args++;
  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_FORM);n_args++;
  XtSetArg(args[n_args],XmNvalue,buf);n_args++;
  XtSetArg(args[n_args],XmNeditable,False);n_args++;
  if(n==0)
    {
      XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_FORM);n_args++;
      XtSetArg(args[n_args],XmNtopOffset,1);n_args++;

    }
  else
    {
      XtSetArg(args[n_args],XmNtopOffset,2);n_args++;
      XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_WIDGET);n_args++;
      XtSetArg(args[n_args],XmNtopWidget,label_eeprom[n-1]);n_args++;
    }
  
  widgetlist[num_wid++]=label_eeprom[n]=
    XtCreateWidget("memory_label",xmTextFieldWidgetClass,dataform_eeprom,
		   args,n_args);

  n_args=t;
  XtSetArg(args[n_args],XmNleftOffset,0);n_args++;
  XtSetArg(args[n_args],XmNcolumns,3);n_args++;
  XtSetArg(args[n_args],XmNmaxLength,2);n_args++;
  XtSetArg(args[n_args],XmNvalue,"0");n_args++;
  t=n_args;

  for(i=0;i<16;i++)
    {
      n_args=t;
      if(i==0) 
	{
	  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_WIDGET);n_args++;
	  XtSetArg(args[n_args],XmNleftWidget,label_eeprom[n]);n_args++;
	}
      else
	{
	  XtSetArg(args[n_args],XmNleftAttachment,XmATTACH_WIDGET);n_args++;
	  XtSetArg(args[n_args],XmNleftWidget,field_eeprom[n*16+i-1]);n_args++;
	}
      
      if(n==0) 
	{XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_FORM);n_args++;}
      else 
	{
	  XtSetArg(args[n_args],XmNtopAttachment,XmATTACH_WIDGET);n_args++;
	  XtSetArg(args[n_args],XmNtopWidget,field_eeprom[(n-1)*16+i]);
	  n_args++;
	}
      widgetlist[num_wid++]=field_eeprom[n*16+i]=
	XtCreateWidget("memory_field",xmTextFieldWidgetClass,dataform_eeprom,
		       args,n_args);
      xs_widget_tip(field_eeprom[n*16+i],eeprom_tip_str[n]);
      XtAddCallback(field_eeprom[n*16+i],XmNfocusCallback,(XtCallbackProc)
		    get_focus_field_eeprom_callback,NULL);
      XtAddCallback(field_eeprom[n*16+i],XmNlosingFocusCallback,
		    (XtCallbackProc)lose_focus_field_eeprom_callback,NULL);
      
    }
  XtVaSetValues(field_eeprom[n*16+i-1],XmNrightAttachment,
		XmATTACH_FORM,NULL);
  *n_wid=num_wid;
}

void resize_eeprom(Widget w,char *m,XConfigureEvent *event,
		 Boolean *continueDispatch)
{
  int n,n_wid;
  Widget *widgetlist;
  
  if(event->type==ConfigureNotify)
    if(event->height!=
       height_min_eeprom_window+nlines_eeprom*eeprom_window_step) 
      {
	nlines_eeprom_old=nlines_eeprom;
	nlines_eeprom=(event->height-height_min_eeprom_window)/
	  eeprom_window_step;
	
	if(nlines_eeprom>controllers[type_micro].size_eeprom/16)
	  {
	    nlines_eeprom=controllers[type_micro].size_eeprom/16;
	    if(nlines_eeprom==nlines_eeprom_old) return;
	  }
	
	if(nlines_eeprom>nlines_eeprom_old)
	  {  /* increase io_lines */
	    
	    if((widgetlist=malloc(nlines_eeprom*17*sizeof(Widget)))==NULL)
	      {
		puts("resize_eeprom(): \
can't allocate memory for 'widgetlist'");
		exit(1);
	      }
	    n_wid=0;
	    for(n=nlines_eeprom_old;n<nlines_eeprom;n++)
	      {
		
		if(label_eeprom[n]==NULL)
		  create_eeprom_line(n,0,widgetlist,&n_wid);
	      }
	    if(n_wid) XtManageChildren(widgetlist,n_wid);
	    free(widgetlist);
	  }
	
	XtVaSetValues(vscr_eeprom_bar,XmNmaximum,
		      controllers[type_micro].size_eeprom/16,XmNsliderSize,
		      nlines_eeprom,NULL);
	
	if(address_eeprom+nlines_eeprom>controllers[type_micro].size_eeprom/16)
	  {
	    address_eeprom=
	      controllers[type_micro].size_eeprom-nlines_eeprom*16;
	    XtVaSetValues(vscr_eeprom_bar,XmNvalue,address_eeprom/16,NULL);
	  }
	full_update_eeprom_window();
      }

}

void full_update_eeprom_window(void)
{
  int n,address;
  char buf[10];
  unsigned char value;
    
  XtVaGetValues(menu[MEMORY_MENU].sub_menu[EEPROM_BUTTON].button,
		XmNset,&value,NULL);
  if(value!=XmSET) return;
  
  for(n=0;n<nlines_eeprom*16;n++)
    {
      if(type_eeprom_window==HEX)
	sprintf(buf,"%X",eeprom_pointer[n+address_eeprom]);
      else sprintf(buf,"%u",eeprom_pointer[n+address_eeprom]);
      XmTextFieldSetString(field_eeprom[n],buf);
      delete_tip(field_eeprom[n],eeprom_tip_str[n]);
      xs_widget_tip(field_eeprom[n],
		    conv_to_bin((int)eeprom_pointer[n+address_eeprom],
				eeprom_tip_str[n]));
    }
  
  for(n=0,address=address_eeprom;n<nlines_eeprom;n++,address+=16)
    {
      sprintf(buf,"%X",address);
      fill_zero(buf,4);
      XmTextFieldSetString(label_eeprom[n],buf);
    }

}

void update_eeprom_window(void)
{
  int n;
  char buf[4];
  unsigned char value;
    
  XtVaGetValues(menu[MEMORY_MENU].sub_menu[EEPROM_BUTTON].button,
		XmNset,&value,NULL);
  if(value!=XmSET) return;
  
  for(n=0;n<nlines_eeprom*16;n++)
    {
      if(type_eeprom_window==HEX)
	sprintf(buf,"%X",eeprom_pointer[n+address_eeprom]);
      else sprintf(buf,"%u",eeprom_pointer[n+address_eeprom]);
      XmTextFieldSetString(field_eeprom[n],buf);
      delete_tip(field_eeprom[n],eeprom_tip_str[n]);
      xs_widget_tip(field_eeprom[n],
		    conv_to_bin((int)eeprom_pointer[n+address_eeprom],
				eeprom_tip_str[n]));
    }
}

void close_button_eeprom(Widget w,XtPointer client_data,XtPointer call_data)
{
  XtVaSetValues(menu[MEMORY_MENU].sub_menu[EEPROM_BUTTON].button,
		XmNset,XmUNSET,NULL);
  XtUnmanageChild(eeprom_window);
}

void switch_button_eeprom(Widget w,XtPointer client_data,XtPointer call_data)
{
  XmString string;
  int n;
  
  if(type_eeprom_window==HEX) 
    {
      type_eeprom_window=DEC;
      string=XmStringGenerate(_("   Hex   "),XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(w,XmNlabelString,string,NULL);
      XmStringFree(string);
      for(n=0;n<nlines_eeprom*16;n++)
	XtVaSetValues(field_eeprom[n],XmNmaxLength,3,NULL);
    }
  else
    {
      type_eeprom_window=HEX;
      string=XmStringGenerate(_(" Decimal "),XmFONTLIST_DEFAULT_TAG,
			      XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
      XtVaSetValues(w,XmNlabelString,string,NULL);
      XmStringFree(string);
      for(n=0;n<nlines_eeprom*16;n++)
	XtVaSetValues(field_eeprom[n],XmNmaxLength,2,NULL);
    }
  update_eeprom_window();  
}

void scroll_bar_eeprom_callback(Widget w,XtPointer client_data,
			       XtPointer call_data)
{
  
  XtVaGetValues(vscr_eeprom_bar,XmNvalue,&address_eeprom,NULL);
  address_eeprom*=16;
  full_update_eeprom_window(); 
}


void get_focus_field_eeprom_callback(Widget w,XtPointer client_data,
				   XtPointer call_data)
{
  XtVaSetValues(w,XmNcursorPositionVisible,True,NULL);
  XtAddCallback(w,XmNvalueChangedCallback,
		(XtCallbackProc)field_eeprom_changed_callback,NULL);
}

void lose_focus_field_eeprom_callback(Widget w,XtPointer client_data,
				    XtPointer call_data)
{
  XtVaSetValues(w,XmNcursorPositionVisible,False,NULL);
  XtRemoveCallback(w,XmNvalueChangedCallback,(XtCallbackProc)
		   field_eeprom_changed_callback,NULL); 
}

void field_eeprom_changed_callback(Widget w,XtPointer client_data,
				   XtPointer call_data)
{
  char *buf,buf2[7];
  int n;
  
  buf=XmTextFieldGetString(w);
  n=0;
  while(n<nlines_eeprom*16&&w!=field_eeprom[n]) n++;
  if(type_eeprom_window==DEC)
    eeprom_pointer[n+address_eeprom]=strtoul(buf,NULL,10);
  else
    {
      sprintf(buf2,"0x%s",buf);
      eeprom_pointer[n+address_eeprom]=strtoul(buf2,NULL,16);
      delete_tip(field_eeprom[n],eeprom_tip_str[n]);
      xs_widget_tip(field_eeprom[n],
		    conv_to_bin((int)eeprom_pointer[n+address_eeprom],
				eeprom_tip_str[n]));
    }
  XtFree(buf);
}

void set_size_eeprom_window(Widget w,XtPointer client_data,XtPointer call_data)
{
  Dimension width,height,height_label;

  XtRemoveCallback(w,XmNpopupCallback,(XtCallbackProc)set_size_eeprom_window,
		   NULL);
  XtVaGetValues(w,XmNwidth,&width,XmNheight,&height,NULL);
  XtVaGetValues(label_eeprom[0],XmNheight,&height_label,NULL);
  eeprom_window_step=height_label+2;
  height_min_eeprom_window=height-nlines_eeprom*eeprom_window_step;

  XtVaSetValues(w,XmNmaxWidth,width,XmNminWidth,width,XmNminHeight,
		height+(4-nlines_eeprom)*eeprom_window_step,
		XmNheightInc,eeprom_window_step,XmNmaxHeight,
		height_min_eeprom_window+MAX_EEPROM_LINES*eeprom_window_step,
		NULL);
}


/* Memory window ----------------------------------------------------------- */
void close_button_memory(Widget w,XtPointer client_data,XtPointer call_data)
{
  XtVaSetValues(menu[MEMORY_MENU].sub_menu[MEMORY_BUTTON].button,
		XmNset,XmUNSET,NULL);
  XtUnmanageChild(memory_window);
}

void refresh_memory_window(Widget w,XtPointer client_data, XtPointer call_data)
{
  XEvent *event;
  Pixmap pixmap;
  
  event=((XmDrawingAreaCallbackStruct *)call_data)->event;
  if(event->type==Expose)  
    {   
      if(event->xexpose.window==XtWindow(mwda.label_widget)) pixmap=mwda.label;
      else 
	if(event->xexpose.window==XtWindow(mwda.address_widget)) 
	  pixmap=mwda.address;
	else
	  if(event->xexpose.window==XtWindow(mwda.data_widget)) 
	    pixmap=mwda.data;
	  else
	    if(event->xexpose.window==XtWindow(mwda.chars_widget))
	      pixmap=mwda.chars;
	    else return;
      XCopyArea(display,pixmap,event->xexpose.window,mwda.gc,event->xexpose.x,
		event->xexpose.y,event->xexpose.width,
		event->xexpose.height,event->xexpose.x,event->xexpose.y);
    };  
}

void draw_memory_window(void)
{
  static char *label=" Address  00  01  02  03  04  05  06  07  08  09  0A  0B  0C  0D  0E  0F       chars";
  char buf[10];
  int n,i;
  GC gc;

  XDrawImageString(display,mwda.label,mwda.label_gc,0,mwda.font_height-1,label,
		   strlen(label));
  
  XFillRectangle(display,mwda.address,mwda.address_gc,0,0,mwda.font_width*8,
		 mwda.font_height*mwda.num_lines+3);
  XFillRectangle(display,mwda.data,mwda.data_gc,
		 0,0,mwda.font_width*64,mwda.font_height*mwda.num_lines+3);
  XFillRectangle(display,mwda.chars,mwda.chars_gc,
		 0,0,mwda.font_width*18,mwda.font_height*mwda.num_lines+3);
  for(n=mwda.cur_line;n<mwda.cur_line+mwda.num_lines&&n<mwda.max_lines;n++)
    {
      sprintf(buf,"%X",n*16);
      fill_zero(buf,6);
      XDrawString(display,mwda.address,mwda.gc,mwda.font_width,
		  mwda.font_height*(n+1-mwda.cur_line),buf,strlen(buf));
      for(i=0;i<16;i++)
	{  
	  switch(mwda.data_type){
	  case HEX: 
	    sprintf(buf,"%X",mwda.pdata[n*16+i]);fill_zero(buf,2);
	    break;
	  case DEC:
	    sprintf(buf,"%u",mwda.pdata[n*16+i]);fill_zero(buf,3);
	    break;
	  }
	  if(mwda.mem_changed[(mwda.mem_offset+n*16+i)/8]&
	     (1<<(mwda.mem_offset+n*16+i)%8)) gc=mwda.gc_new;
	  else gc=mwda.gc;

	  XDrawString(display,mwda.data,gc,mwda.font_width*(4*i+1),
		      mwda.font_height*(n+1-mwda.cur_line),buf,strlen(buf));
	  XDrawString(display,mwda.chars,gc,mwda.font_width*(i+1),
		      mwda.font_height*(n+1-mwda.cur_line),
		      &mwda.pdata[n*16+i],1);
	}
    }
}

void full_update_memory_window(void)
{
  if(!memory_window) return;
  if(!XtIsManaged(memory_window)) return;
  draw_memory_window();
  XCopyArea(display,mwda.label,XtWindow(mwda.label_widget),mwda.label_gc,
	    0,0,mwda.width,mwda.font_height+3,0,0);
  XCopyArea(display,mwda.address,XtWindow(mwda.address_widget),mwda.address_gc,
	    0,0,mwda.font_width*8,mwda.font_height*mwda.num_lines+3,0,0);
  XCopyArea(display,mwda.data,XtWindow(mwda.data_widget),mwda.data_gc,
	    0,0,mwda.font_width*64,mwda.font_height*mwda.num_lines+3,0,0);
  XCopyArea(display,mwda.chars,XtWindow(mwda.chars_widget),mwda.chars_gc,
	    0,0,mwda.font_width*18,mwda.font_height*mwda.num_lines+3,0,0);
}

void memory_scroll_bar_callback(Widget w,XtPointer client_data,
				XtPointer call_data)
{
  char buf[10];

  XtVaGetValues(mwda.vscr_bar,XmNvalue,&mwda.cur_line,NULL);
  draw_memory_window();
  full_update_memory_window();
  
  sprintf(buf,"%X",mwda.cur_line*16);
  fill_zero(buf,6);
  XtVaSetValues(mwda.address_text_widget,XmNvalue,buf,NULL);
}

void memory_resize_callback(Widget w,XtPointer client_data,
			    XmDrawingAreaCallbackStruct *call_data)
{
  Dimension height;
  if(call_data->reason!=XmCR_RESIZE) return;

  XtVaGetValues(mwda.address_widget,XmNheight,&height,NULL);
  mwda.num_lines=height/mwda.font_height;
  printf("New Height: %d\tlines: %d\n",height,mwda.num_lines);


  XFreePixmap(display,mwda.label);
  XFreePixmap(display,mwda.address);
  XFreePixmap(display,mwda.data);
  XFreePixmap(display,mwda.chars);
  mwda.label=CreatePixmap(XtWindow(toplevel),mwda.width,
    			  mwda.font_height+3,mwda.label_bg);
  mwda.address=CreatePixmap(XtWindow(toplevel),mwda.font_width*8,
    			    mwda.font_height*mwda.num_lines+3,mwda.address_bg);
  mwda.data=CreatePixmap(XtWindow(toplevel),mwda.font_width*64,
    			 mwda.font_height*mwda.num_lines+3,mwda.data_bg);
  mwda.chars=CreatePixmap(XtWindow(toplevel),mwda.font_width*18,
    			  mwda.font_height*mwda.num_lines+3,mwda.chars_bg); 
  draw_memory_window();
}

void init_memory_window(void)
{
  if(!memory_window) return;

  switch(mwda.memory_type){
  case M_REGISTERS:
    mwda.pdata=gpr_pointer;
    mwda.max_lines=controllers[type_micro].size_gpr/16;
    mwda.mem_changed=mem_file_changed;
    mwda.mem_offset=0;
    break;
  case M_IO_REGISTERS:
    mwda.pdata=io_pointer;
    mwda.max_lines=controllers[type_micro].size_io_reg/16;
    mwda.mem_changed=mem_file_changed;
    mwda.mem_offset=controllers[type_micro].size_gpr;
    break;
  case M_EX_IO_REGISTERS:
    break;
  case M_SRAM:
    mwda.pdata=sram_pointer;
    mwda.max_lines=controllers[type_micro].size_sram/16;
    mwda.mem_changed=mem_file_changed;
    mwda.mem_offset=controllers[type_micro].size_io_reg+
      controllers[type_micro].size_gpr;
    break;
  case M_DATA:
    mwda.pdata=gpr_pointer;
    mwda.max_lines=(controllers[type_micro].size_gpr+
		    controllers[type_micro].size_io_reg+
		    controllers[type_micro].size_sram)/16;
    mwda.mem_changed=mem_file_changed;
    mwda.mem_offset=0;
    break;    
  case M_EEPROM:
    mwda.pdata=eeprom_pointer;
    mwda.max_lines=controllers[type_micro].size_eeprom/16;
    mwda.mem_changed=eeprom_changed;
    mwda.mem_offset=0;
    break;
  }
  mwda.cur_line=0;

  if(mwda.num_lines>mwda.max_lines)
    {
      if(mwda.max_lines)
	XtVaSetValues(mwda.vscr_bar,XmNsliderSize,mwda.max_lines,
		      XmNmaximum,mwda.max_lines,NULL);
      else XtVaSetValues(mwda.vscr_bar,XmNsliderSize,1,
			 XmNmaximum,1,NULL);
    }
  else
    XtVaSetValues(mwda.vscr_bar,XmNsliderSize,mwda.num_lines,
		  XmNmaximum,mwda.max_lines,NULL);  
}

void combo_mem_callback(Widget w,XtPointer client_data,
			XmComboBoxCallbackStruct *call_data)
{
  mwda.memory_type=call_data->item_position;
  init_memory_window();
  draw_memory_window();
  full_update_memory_window();
}

void switch_memory_data_type_callback(Widget w,XtPointer client_data,
				      XtPointer call_data)
{
  XmString string;

  if(mwda.data_type==HEX)
    {
      mwda.data_type=DEC;
      string=XmStringCreateLtoR("DEC",XmSTRING_DEFAULT_CHARSET);
      XtVaSetValues(w,XmNlabelString,string,NULL);
      XmStringFree(string);
    }
  else
    {
      mwda.data_type=HEX;
      string=XmStringCreateLtoR("HEX",XmSTRING_DEFAULT_CHARSET);
      XtVaSetValues(w,XmNlabelString,string,NULL);
      XmStringFree(string);
    }
  draw_memory_window();
  full_update_memory_window();
}

void create_memory_window(void)
{
  Widget shell,memory_frame,button,scroll,memory_row,combo_mem;
  XmString string,*list;
  XFontStruct *font_struct;
  XGCValues values;
  XColor color;
  int size_list,n,size_string,t;

  /* Create widgets */  
  shell=XtVaCreateManagedWidget("Memory_Shell",xmDialogShellWidgetClass,
				toplevel,XmNtitle,_("Memory"),
				XmNallowShellResize,True,NULL);
  memory_window=XtVaCreateManagedWidget("memory_form",xmFormWidgetClass,shell,
					XmNautoUnmanage,False,
					XmNdialogStyle,XmDIALOG_MODELESS,
					XmNfractionBase,3,NULL);
  memory_row=
    XtVaCreateManagedWidget("memory_row",xmRowColumnWidgetClass,memory_window,
			    XmNtopAttachment,XmATTACH_FORM,XmNtopOffset,1,
			    XmNorientation,XmHORIZONTAL,NULL);

  size_string=1;n=0;
  for(n=0;n<XtNumber(memory_type_str);n++)
    if((t=strlen(memory_type_str[n]))>size_string) size_string=t;
  
  size_list=n;  
  if((list=malloc(size_list*sizeof(XmString)))==NULL)
    {
      puts("create_memory_window(): can't allocate memory for 'list'");
      exit(1);
    }

  for(n=0;n<size_list;n++)
    list[n]=XmStringCreateLtoR(gettext(memory_type_str[n]),
			       XmSTRING_DEFAULT_CHARSET);
  
  combo_mem=XtVaCreateManagedWidget("memory_type_box",xmComboBoxWidgetClass,
				    memory_row,
				    XmNcomboBoxType,XmDROP_DOWN_LIST,
				    XmNitems,list,XmNitemCount,size_list,
				    XmNvisibleItemCount,size_list,
				    XmNcolumns,size_string,
				    XmNselectedPosition,M_DATA,NULL);
  XtVaSetValues(XtNameToWidget(combo_mem,"*Text"),XmNshadowThickness,1,NULL);
  for(n=0;n<size_list;n++) XmStringFree(list[n]);
  free(list);
  
  XtAddCallback(combo_mem,XmNselectionCallback,
		(XtCallbackProc)combo_mem_callback,NULL);

  string=XmStringGenerate(_("Address:"),XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  XtVaCreateManagedWidget("address_label",xmLabelGadgetClass,memory_row,
			  XmNlabelString,string,NULL);
  XmStringFree(string);

  mwda.address_text_widget=
    XtVaCreateManagedWidget("address_textfield",xmTextFieldWidgetClass,
			    memory_row,XmNcolumns,7,XmNvalue,"000000",NULL);
  
  string=XmStringGenerate(_("Find:"),XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  XtVaCreateManagedWidget("find_label",xmLabelGadgetClass,memory_row,
			  XmNlabelString,string,NULL);
  XmStringFree(string);

  XtVaCreateManagedWidget("find_textfield",xmTextFieldWidgetClass,
			  memory_row,XmNcolumns,10,NULL);
  
  mwda.data_type_widget=
    XtVaCreateManagedWidget("HEX",xmPushButtonWidgetClass,
			    memory_row,NULL);
  XtAddCallback(mwda.data_type_widget,XmNactivateCallback,
		(XtCallbackProc)switch_memory_data_type_callback,NULL);

  memory_frame=
    XtVaCreateManagedWidget("memory_frame",xmFrameWidgetClass,memory_window,
			    XmNtopAttachment,XmATTACH_WIDGET,
			    XmNtopWidget,memory_row,
			    XmNleftAttachment,XmATTACH_FORM,XmNleftOffset,5,
			    XmNrightAttachment,XmATTACH_FORM,
			    XmNrightOffset,5,XmNmarginWidth,2,
			    XmNmarginHeight,2,NULL);  

  /* create scroll window */
  scroll=
    XtVaCreateManagedWidget("Memory_Scroll",xmScrolledWindowWidgetClass,
			    memory_frame,
			    XmNscrollingPolicy,XmAPPLICATION_DEFINED,
			    XmNscrollBarDisplayPolicy,XmSTATIC,
			    XmNleftAttachment,XmATTACH_FORM,XmNleftOffset,3,
			    XmNrightAttachment,XmATTACH_FORM,XmNrightOffset,2,
			    XmNtopAttachment,XmATTACH_WIDGET,XmNtopWidget,
			    memory_frame,XmNvisualPolicy,XmVARIABLE,
			    XmNbottomAttachment,XmATTACH_FORM,
			    XmNbottomOffset,3,NULL);
  mwda.draw=XtVaCreateManagedWidget("DrawingManager",
				    xmFormWidgetClass,scroll,
				    XmNbackground,
				    WhitePixel(display,scr_num),
				    XmNforeground,
				    BlackPixel(display,scr_num),NULL); 
  mwda.label_widget=
    XtVaCreateManagedWidget("Drawing area",xmDrawingAreaWidgetClass,mwda.draw,
			    XmNborderWidth,1,NULL);
  mwda.address_widget=
    XtVaCreateManagedWidget("Drawing area",xmDrawingAreaWidgetClass,mwda.draw,
			    XmNborderWidth,1,XmNtopAttachment,XmATTACH_WIDGET,
			    XmNtopWidget,mwda.label_widget,
			    XmNbottomAttachment,XmATTACH_FORM,NULL);
  mwda.data_widget=
    XtVaCreateManagedWidget("Drawing area",xmDrawingAreaWidgetClass,mwda.draw,
			    XmNborderWidth,1,XmNtopAttachment,XmATTACH_WIDGET,
			    XmNtopWidget,mwda.label_widget,
			    XmNbottomAttachment,XmATTACH_FORM,NULL);
  mwda.chars_widget=
    XtVaCreateManagedWidget("Drawing area",xmDrawingAreaWidgetClass,mwda.draw,
			    XmNborderWidth,1,XmNtopAttachment,XmATTACH_WIDGET,
			    XmNtopWidget,mwda.label_widget,
			    XmNbottomAttachment,XmATTACH_FORM,NULL);
  mwda.vscr_bar=
    XtVaCreateManagedWidget("memory_scr_bar",xmScrollBarWidgetClass,scroll,
			    XmNorientation,XmVERTICAL,XmNincrement,1,
			    XmNminimum,0,XmNvalue,0,NULL);

  XtAddCallback(mwda.vscr_bar,XmNvalueChangedCallback,
		(XtCallbackProc)memory_scroll_bar_callback,NULL);
  XtAddCallback(mwda.vscr_bar,XmNdragCallback,
		(XtCallbackProc)memory_scroll_bar_callback,NULL);
  XmScrolledWindowSetAreas(scroll,NULL,mwda.vscr_bar,mwda.draw);
  
  string=XmStringGenerate(_("Close"),XmFONTLIST_DEFAULT_TAG,
			  XmCHARSET_TEXT,_MOTIF_DEFAULT_LOCALE);
  button=XtVaCreateManagedWidget("Close",xmPushButtonGadgetClass,
				 memory_window,
				 XmNleftAttachment,XmATTACH_POSITION,
				 XmNleftPosition,1,
				 XmNrightAttachment,XmATTACH_POSITION,
				 XmNrightPosition,2,XmNlabelString,string,
				 XmNbottomAttachment,XmATTACH_FORM,
				 XmNbottomOffset,6,NULL);
  
  XtAddCallback(button,
		XmNactivateCallback,(XtCallbackProc)close_button_memory,NULL);
  XmStringFree(string);
  XtVaSetValues(memory_frame,XmNbottomOffset,6,XmNbottomAttachment,
		XmATTACH_WIDGET,XmNbottomWidget,button,NULL);
  /*  XtAddCallback(shell,XmNpopupCallback,(XtCallbackProc)set_size_eeprom_window, */
  /*  	       NULL); */
  /*   XtAddEventHandler(shell,StructureNotifyMask,False,resize_eeprom,NULL); */

  /* init window */
  mwda.num_lines=20;
  mwda.cur_line=0;
  mwda.data_type=HEX;
  mwda.pdata=gpr_pointer;
  mwda.memory_type=M_DATA;
  mwda.mem_changed=mem_file_changed;
  mwda.mem_offset=0;
  mwda.max_lines=(controllers[type_micro].size_gpr+
		  controllers[type_micro].size_io_reg+
		  controllers[type_micro].size_sram)/16;
  if(mwda.num_lines>mwda.max_lines)
    XtVaSetValues(mwda.vscr_bar,XmNsliderSize,mwda.max_lines,
		  XmNmaximum,mwda.max_lines,NULL);
  else
    XtVaSetValues(mwda.vscr_bar,XmNsliderSize,mwda.num_lines,
		  XmNmaximum,mwda.max_lines,NULL);

  values.foreground=BlackPixel(display,scr_num);
  values.background=WhitePixel(display,scr_num);
  mwda.gc=XCreateGC(display,XtWindow(toplevel),
		    GCForeground|GCBackground,&values);
  
  if((font_struct=XQueryFont(display,XGContextFromGC(mwda.gc)))==NULL)
    {
      fprintf(stderr,"Can't load font\n");
      exit(1);
    }
  mwda.font_width=font_struct->max_bounds.width;
  mwda.font_height=font_struct->ascent+font_struct->descent;
  XFreeFontInfo(NULL,font_struct,1);
  mwda.width=92*mwda.font_width;
  mwda.height=(mwda.num_lines+3)*mwda.font_height;

 /* allocate color */ 
  if(!XAllocNamedColor(display,DefaultColormap(display,scr_num), 
                       "rgb:A0/A0/A0",&color,&color))
    {                                             
      puts("Can't allocate color.");
      color.pixel=WhitePixel(display,scr_num);
    }
  values.background=mwda.label_bg=color.pixel;
  mwda.label_gc=XCreateGC(display,XtWindow(toplevel),GCForeground|GCBackground,
			  &values);
  if(!XAllocNamedColor(display,DefaultColormap(display,scr_num), 
                       "rgb:80/80/D0",&color,&color))
    {                                             
      puts("Can't allocate color.");
      color.pixel=WhitePixel(display,scr_num);
    }
  values.foreground=mwda.address_bg=color.pixel;
  mwda.address_gc=
    XCreateGC(display,XtWindow(toplevel),GCForeground,&values);
  if(!XAllocNamedColor(display,DefaultColormap(display,scr_num), 
                       "rgb:F0/F0/F0",&color,&color))
    {                                             
      puts("Can't allocate color.");
      color.pixel=WhitePixel(display,scr_num);
    }
  values.foreground=mwda.data_bg=mwda.chars_bg=color.pixel;
  mwda.chars_gc=mwda.data_gc=
    XCreateGC(display,XtWindow(toplevel),GCForeground,&values);

  if(!XAllocNamedColor(display,DefaultColormap(display,scr_num), 
                       "red",&color,&color))
    {                                             
      puts("Can't allocate color.");
      color.pixel=BlackPixel(display,scr_num);
    }
  values.foreground=color.pixel;
  mwda.gc_new=XCreateGC(display,XtWindow(toplevel),GCForeground,&values);


  XtVaSetValues(mwda.label_widget,XmNwidth,mwda.width,XmNheight,
		mwda.font_height+3,XmNbackground,mwda.label_bg,NULL);
  XtVaSetValues(mwda.address_widget,XmNwidth,mwda.font_width*8,XmNheight,
		mwda.font_height*mwda.num_lines,XmNx,0,
		XmNleftAttachment,XmATTACH_NONE,
		XmNbottomOffset,mwda.font_height/2,
		XmNtopOffset,mwda.font_height/2,
		XmNbackground,mwda.address_bg,NULL);
  XtVaSetValues(mwda.data_widget,XmNwidth,mwda.font_width*64,
		XmNheight,mwda.font_height*mwda.num_lines,
		XmNleftAttachment,XmATTACH_NONE,
		XmNx,mwda.font_width*9,
		XmNbottomOffset,mwda.font_height/2,
		XmNtopOffset,mwda.font_height/2,
		XmNbackground,mwda.data_bg,NULL);
  XtVaSetValues(mwda.chars_widget,XmNwidth,mwda.font_width*18,
		XmNheight,mwda.font_height*mwda.num_lines,
		XmNleftAttachment,XmATTACH_NONE,
		XmNx,mwda.font_width*74,
		XmNbottomOffset,mwda.font_height/2,
		XmNtopOffset,mwda.font_height/2,
		XmNbackground,mwda.chars_bg,NULL);


  mwda.label=CreatePixmap(XtWindow(toplevel),mwda.width,
			  mwda.font_height+3,mwda.label_bg);
  mwda.address=CreatePixmap(XtWindow(toplevel),mwda.font_width*8,
			    mwda.font_height*mwda.num_lines+3,mwda.address_bg);
  mwda.data=CreatePixmap(XtWindow(toplevel),mwda.font_width*64,
			 mwda.font_height*mwda.num_lines+3,mwda.data_bg);
  mwda.chars=CreatePixmap(XtWindow(toplevel),mwda.font_width*18,
			  mwda.font_height*mwda.num_lines+3,mwda.chars_bg);
  XtAddCallback(mwda.label_widget,XmNexposeCallback,
		(XtCallbackProc)refresh_memory_window,NULL);
  XtAddCallback(mwda.address_widget,XmNexposeCallback,
		(XtCallbackProc)refresh_memory_window,NULL);
  XtAddCallback(mwda.data_widget,XmNexposeCallback,
		(XtCallbackProc)refresh_memory_window,NULL);
  XtAddCallback(mwda.chars_widget,XmNexposeCallback,
		(XtCallbackProc)refresh_memory_window,NULL);
  XtAddCallback(mwda.address_widget,XmNresizeCallback, 
  		(XtCallbackProc)memory_resize_callback,NULL);

  draw_memory_window();
  XtVaSetValues(shell,XmNheightInc,mwda.font_height,NULL);
}
